/*
 * Dan Aloni <da-x@colinux.org>, 2005 (c)
 *
 * The code is licensed under the GPL. See the COPYING file at
 * the root directory.
 *
 * Partially based on work from Linux's fs/pipe.c.
 */

#include "linux_inc.h"
#include <linux/mman.h>

#include <linux/upm_fs_i.h>
#include "interface.h"

#include <asm/uaccess.h>
#include <asm/ioctls.h>

static void free_upm_info(struct inode *inode);

static int upm_read_release(struct inode *inode, struct file *filp)
{ 
	if (filp)
		free_upm_info(inode);

	return 0;
}

static int upm_read_open(struct inode *inode, struct file *filp)
{
	return 0;
}

struct file_operations read_upm_fops = {
        .llseek         = generic_file_llseek,
        .read           = generic_file_read,
        .mmap           = generic_file_readonly_mmap,
        .sendfile       = generic_file_sendfile,
	.open		= upm_read_open,
	.release	= upm_read_release,
};

static int upm_readpage(struct file *file, struct page *page)
{
	struct upm_inode_info *info;
	unsigned char *pageptr;
	unsigned long readsize;
	loff_t inodesize;
	loff_t readoffset;
	int error = 0;

        if (PageUptodate(page))
                goto out;

	info = UPM_INFO(file->f_dentry->d_inode);

	pageptr = kmap(page);
	memset(pageptr, 0, PAGE_SIZE);
	readsize = PAGE_SIZE;
	inodesize = i_size_read(file->f_dentry->d_inode);
	readoffset = page->index;
	readoffset = readoffset*PAGE_SIZE + info->offset;
	if (readoffset + readsize > inodesize)
		readsize = inodesize - readoffset;
	error = kernel_read(info->wrapped_file, page->index*PAGE_SIZE + info->offset, pageptr, readsize);
	kunmap(page);

	SetPageUptodate(page);
	unlock_page(page);

out:
        return error;
}

static struct address_space_operations read_upm_aops = {
	.readpage = upm_readpage,
};

static void free_upm_info(struct inode *inode)
{
	struct upm_inode_info *info = UPM_INFO(inode);
	fput(info->wrapped_file);
	kfree(info);
}

struct inode* upm_new(struct inode* inode, unsigned long offset)
{
	struct upm_inode_info *info;

	info = kmalloc(sizeof(struct upm_inode_info), GFP_KERNEL);
	if (!info)
		goto fail_page;

	memset(info, 0, sizeof(*info));

	info->offset = offset;
	inode->i_pipe = (typeof(inode->i_pipe))info;

	return inode;

fail_page:
	return NULL;
}

static struct vfsmount *upm_mnt;

static int upmfs_delete_dentry(struct dentry *dentry)
{
	module_put(THIS_MODULE);

	return 1;
}

static struct dentry_operations upmfs_dentry_operations = {
	.d_delete	= upmfs_delete_dentry,
};


static struct inode *get_upm_inode(unsigned long offset)
{
	struct inode *inode = new_inode(upm_mnt->mnt_sb);

	if (!inode)
		goto fail_inode;

	if(!upm_new(inode, offset))
		goto fail_iput;

	inode->i_fop = &read_upm_fops;
	inode->i_mapping->a_ops = &read_upm_aops;

	/*
	 * Mark the inode dirty from the very beginning,
	 * that way it will never be moved to the dirty
	 * list because "mark_inode_dirty()" will think
	 * that it already _is_ on the dirty list.
	 */
	inode->i_state = I_DIRTY;
	inode->i_mode = S_IFIFO | S_IRUSR;
	inode->i_uid = current->fsuid;
	inode->i_gid = current->fsgid;
	inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	inode->i_blksize = PAGE_SIZE;

	return inode;

fail_iput:
	iput(inode);
fail_inode:
	return NULL;
}

struct file *upm_get_unaligned_map(struct file *wrapped_file, unsigned long offset)
{
	struct qstr this;
	struct dentry *dentry;
	struct inode * inode;
	struct file *filp;
	struct upm_inode_info *info;
	unsigned long name_size;
	char *name, *name_ret;
	int error;

	error = -ENFILE;
	filp = get_empty_filp();
	if (!filp)
		goto no_files;

	error = -ENOMEM;
	inode = get_upm_inode(offset);
	if (!inode)
		goto close_file;

	name_size = 0x100;
	name = kmalloc(name_size, GFP_KERNEL);
	if (!name)
		goto close_inode;

	error = -ENOMEM;
	name_ret = d_path(wrapped_file->f_dentry, wrapped_file->f_vfsmnt, name, name_size);
	if (IS_ERR(name_ret)) {
		error = PTR_ERR(name_ret);
		goto free_name;
	}

	memmove(name, name_ret, strlen(name_ret)+1);

#if (0)
	{
		unsigned long cur_size;
		/* show more info in /proc/pid/maps */
		cur_size = strlen(name);
		snprintf(name + cur_size, name_size - cur_size, " upmfs:[%lu]:%04x", inode->i_ino, offset);
	}
#endif

	this.name = name;
	this.len = strlen(name);
	this.hash = inode->i_ino; /* will go */

	info = UPM_INFO(inode);
	info->wrapped_file = wrapped_file;
	get_file(wrapped_file);

	error = -ENOMEM;
	dentry = d_alloc(upm_mnt->mnt_sb->s_root, &this);
	if (!dentry)
		goto free_name;

	if (!try_module_get(THIS_MODULE))
		goto free_name;

	inode->i_size = wrapped_file->f_dentry->d_inode->i_size;
	inode->i_uid = wrapped_file->f_dentry->d_inode->i_uid;
	inode->i_gid = wrapped_file->f_dentry->d_inode->i_gid;
	inode->i_atime = wrapped_file->f_dentry->d_inode->i_atime;
	inode->i_mtime = wrapped_file->f_dentry->d_inode->i_mtime;
	inode->i_ctime = wrapped_file->f_dentry->d_inode->i_ctime;

	dentry->d_op = &upmfs_dentry_operations;
	d_add(dentry, inode);

	filp->f_vfsmnt = mntget(upm_mnt);
	filp->f_dentry = dentry;
	filp->f_mapping = inode->i_mapping;

	/* read file */
	filp->f_pos = 0;
	filp->f_flags = O_RDONLY;
	filp->f_op = &read_upm_fops;
	filp->f_mode = FMODE_READ;
	filp->f_version = 0;
	kfree(name);

	return filp;

free_name:
	kfree(name);

close_inode:
	free_upm_info(inode);
	iput(inode);

close_file:
	fput(filp);

no_files:
	return ERR_PTR(error);
}

EXPORT_SYMBOL(upm_get_unaligned_map);

/*
 * upmfs should _never_ be mounted by userland - too much of security hassle,
 * no real gain from having the whole whorehouse mounted. So we don't need
 * any operations on the root directory. However, we need a non-trivial
 * d_name - upm: will go nicely and kill the special-casing in procfs.
 */

static struct super_block *upmfs_get_sb(struct file_system_type *fs_type,
					int flags, const char *dev_name, void *data)
{
	/* TODO: replace this so /proc/self/maps is more readable */
	return get_sb_pseudo(fs_type, "", NULL, UPMFS_MAGIC);
}

static struct file_system_type upmfs_type = {
	.name		= "upmfs",
	.get_sb		= upmfs_get_sb,
	.kill_sb	= kill_anon_super,
};


static int __init init_upmfs(void)
{	
	int err = register_filesystem(&upmfs_type);
	if (err)
		goto exit;

	upm_mnt = kern_mount(&upmfs_type);
	if (IS_ERR(upm_mnt))
		goto exit_2;

	err = upmfs_interface_init();

	return err;

exit_2:
	err = PTR_ERR(upm_mnt);
	unregister_filesystem(&upmfs_type);

exit:
	return err;
}

static void __exit exit_upmfs(void)
{
	unregister_filesystem(&upmfs_type);
	mntput(upm_mnt);
	upmfs_interface_exit();
}

fs_initcall(init_upmfs);
module_exit(exit_upmfs);

MODULE_LICENSE("GPL");
