/*
 * Dan Aloni <da-x@colinux.org>, 2005 (c)
 *
 * The code is licensed under the GPL. See the COPYING file at
 * the root directory.
 */

#include "linux_inc.h"

#include <linux/proc_fs.h>
#include <linux/upm_fs_i.h>

#include "interface.h"

static struct proc_dir_entry *upmfs_proc_root;
static struct proc_dir_entry *upmfs_proc_ioctl;

int upmfs_func_proc_open(struct inode *inode, struct file *file)
{
        if (!try_module_get(THIS_MODULE))
		return -EBUSY;

	return 0;
}

int upmfs_ioctl_unaligned_map(struct upm_io_unaligned_map __user *arg)
{
	struct upm_io_unaligned_map io;
	struct file *new_file;
	struct file *old_file;
	int fd, new_fd;
	int error;

	/* get parameters from userspace */
	if (unlikely(copy_from_user(&io, arg, sizeof(io))))
		return -EFAULT;

	/* access 'struct file *' for the user-provided file descriptor */
	fd = io.old_fd;
	old_file = fcheck(fd);
	if (unlikely(!old_file))
		return -EIO;
		
	/* allocate a new file descriptor */
	new_fd = get_unused_fd();
	if (unlikely(new_fd < 0)) {
		error = new_fd;
		goto exit_error;
	}

	/* create a unaligned mapping 'struct file *' wrapper */
	new_file = upm_get_unaligned_map(old_file, io.offset);
	if (unlikely(IS_ERR(new_file))) {
		error = PTR_ERR(new_file);
		goto exit_error_fd;
	}

	io.new_fd = new_fd;

	/* return result to userspace */
	if (unlikely(copy_to_user(arg, &io, sizeof(io)))) {
		error = -EFAULT;
		goto exit_error_file;
	}

	fd_install(new_fd, new_file);
	
	return 0;


/* error path */
exit_error_file:
	fput(new_file);

exit_error_fd:
	put_unused_fd(new_fd);

exit_error:
	return error;
}

int upmfs_func_proc_ioctl(struct inode *inode, struct file *file, 
			  unsigned int cmd, unsigned long arg)
{
	int error = -EINVAL;

	switch (cmd) { 
	case UPM_IO_UNALIGNED_MAP:
		error = upmfs_ioctl_unaligned_map((struct upm_io_unaligned_map __user *)arg);
		break;
	}

	return error;
}

ssize_t upmfs_func_proc_read(struct file *file, char __user *buffer, size_t size, loff_t *poffset)
{
	return -EIO;
}

ssize_t upmfs_func_proc_write(struct file *file, const char __user *buffer, size_t size, loff_t *poffset)
{
	return -EIO;
}

int upmfs_func_proc_release(struct inode *inode, struct file *file)
{
	module_put(THIS_MODULE);

	return 0;
}

static struct file_operations upmfs_proc_fileops = {
        .open           = upmfs_func_proc_open,
	.ioctl          = upmfs_func_proc_ioctl,
        .read           = upmfs_func_proc_read,
	.write          = upmfs_func_proc_write,
        .release        = upmfs_func_proc_release,
};

int upmfs_interface_init(void)
{
	int retval = 0;

	retval = -EIO;
	upmfs_proc_root = proc_mkdir("upmfs", &proc_root);
	if (upmfs_proc_root == NULL) {
		goto error;
	}
	upmfs_proc_root->owner = THIS_MODULE;
	
	upmfs_proc_ioctl = create_proc_entry("ioctl", S_IFREG|S_IRUSR|S_IWUSR, upmfs_proc_root);
	if (!upmfs_proc_ioctl) {
		goto error_root;
	}

	upmfs_proc_ioctl->proc_fops = &upmfs_proc_fileops;
	upmfs_proc_ioctl->owner = THIS_MODULE;
	retval = 0;

	return retval;

error_root:
	remove_proc_entry("upmfs", &proc_root);

error:
	return retval;
}

void upmfs_interface_exit(void)
{
	remove_proc_entry("ioctl", upmfs_proc_root);
	remove_proc_entry("upmfs", &proc_root);
}
